@@ -22,30 +22,46 @@ from pprint import pprint
 from cStringIO import StringIO
 import json , urllib2, httplib, socket, ssl
 
-API_VERSION = '1.9.0'
+API_VERSION = '1.10.2'
 
+class HTTPSConnectionChain(httplib.HTTPSConnection):
+    _preferred_ssl_protos = (
+        ('TLSv1' , ssl.PROTOCOL_TLSv1) , 
+        ('SSLv3' , ssl.PROTOCOL_SSLv3) ,
+        ('SSLv23' , ssl.PROTOCOL_SSLv23) ,
+    )
+    _ssl_working_proto = None
 
-class HTTPSConnectionV3(httplib.HTTPSConnection):
-    def __init__(self, *args, **kwargs):
-        httplib.HTTPSConnection.__init__(self, *args, **kwargs)
-        
     def connect(self):
         sock = socket.create_connection((self.host, self.port), self.timeout)
         if self._tunnel_host:
             self.sock = sock
             self._tunnel()
-        try:
-            self.sock = ssl.wrap_socket(sock, self.key_file, self.cert_file, ssl_version=ssl.PROTOCOL_SSLv3)
-        except ssl.SSLError, e:
-            print("Trying SSLv3.")
-            self.sock = ssl.wrap_socket(sock, self.key_file, self.cert_file, ssl_version=ssl.PROTOCOL_SSLv23)           
+        if self._ssl_working_proto is not None:
+            # If we have a working proto, let's use that straight away
+            print 'working proto: %s' % self._ssl_working_proto
+            self.sock = ssl.wrap_socket(sock, self.key_file, self.cert_file,
+                ssl_version=self._ssl_working_proto)
+            return
+        # Try connecting via the different SSL protos in preference order
+        for proto_name , proto in self._preferred_ssl_protos:
+            try:
+                self.sock = ssl.wrap_socket(sock, self.key_file, self.cert_file,
+                    ssl_version=proto)
+            except:
+                pass
+            else:
+                # Cache the working ssl version
+                HTTPSConnectionChain._ssl_working_proto = proto
+                break
+
+
+class HTTPSHandlerChain(urllib2.HTTPSHandler):
+    def https_open(self , req):
+        return self.do_open(HTTPSConnectionChain, req)
 
-class HTTPSHandlerV3(urllib2.HTTPSHandler):
-    def https_open(self, req):
-        return self.do_open(HTTPSConnectionV3, req)
 # install opener
-urllib2.install_opener(urllib2.build_opener(HTTPSHandlerV3()))
-
+urllib2.install_opener(urllib2.build_opener(HTTPSHandlerChain()))
 
 class PysHTTPRedirectHandler(urllib2.HTTPRedirectHandler):
     """
@@ -87,7 +103,7 @@ class Connection(object):
         serverPath:str      The base resource path for the subsonic views.
                             This is useful if you have your subsonic server
                             behind a proxy and the path that you are proxying
-                            is differnt from the default of '/rest'.
+                            is different from the default of '/rest'.
                             Ex: 
                                 serverPath='/path/to/subs'
                                 
@@ -834,11 +850,47 @@ class Connection(object):
         self._checkStatus(res)
         return res
 
-    def createUser(self , username , password , ldapAuthenticated=False ,
-            adminRole=False , settingsRole=True , streamRole=True ,
-            jukeboxRole=False , downloadRole=False , uploadRole=False ,
-            playlistRole=False , coverArtRole=False , commentRole=False ,
-            podcastRole=False , shareRole=False):
+    def getUsers(self):
+        """
+        since 1.8.0
+
+        Gets a list of users
+
+        returns a dict like the following
+
+        {u'status': u'ok',
+         u'users': {u'user': [{u'adminRole': True,
+                   u'commentRole': True,
+                   u'coverArtRole': True,
+                   u'downloadRole': True,
+                   u'jukeboxRole': True,
+                   u'playlistRole': True,
+                   u'podcastRole': True,
+                   u'scrobblingEnabled': True,
+                   u'settingsRole': True,
+                   u'shareRole': True,
+                   u'streamRole': True,
+                   u'uploadRole': True,
+                   u'username': u'user1'},
+                   ...
+                   ...
+                   ]} ,
+         u'version': u'1.10.2',
+         u'xmlns': u'http://subsonic.org/restapi'}
+        """
+        methodName = 'getUsers'
+        viewName = '%s.view' % methodName
+
+        req = self._getRequest(viewName)
+        res = self._doInfoReq(req)
+        self._checkStatus(res)
+        return res
+
+    def createUser(self , username , password , email , 
+            ldapAuthenticated=False , adminRole=False , settingsRole=True , 
+            streamRole=True , jukeboxRole=False , downloadRole=False , 
+            uploadRole=False , playlistRole=False , coverArtRole=False , 
+            commentRole=False , podcastRole=False , shareRole=False):
         """
         since: 1.1.0
 
@@ -847,6 +899,7 @@ class Connection(object):
 
         username:str        The username of the new user
         password:str        The password for the new user
+        email:str           The email of the new user
         <For info on the boolean roles, see http://subsonic.org for more info>
 
         Returns a dict like the following:
@@ -859,7 +912,7 @@ class Connection(object):
         viewName = '%s.view' % methodName
         hexPass = 'enc:%s' % self._hexEnc(password)
 
-        q = {'username': username , 'password': hexPass ,
+        q = {'username': username , 'password': hexPass , 'email': email ,
             'ldapAuthenticated': ldapAuthenticated , 'adminRole': adminRole ,
             'settingsRole': settingsRole , 'streamRole': streamRole ,
             'jukeboxRole': jukeboxRole , 'downloadRole': downloadRole ,
@@ -872,6 +925,44 @@ class Connection(object):
         self._checkStatus(res)
         return res
 
+    def updateUser(self , username ,  password=None , email=None ,
+            ldapAuthenticated=False , adminRole=False , settingsRole=True , 
+            streamRole=True , jukeboxRole=False , downloadRole=False , 
+            uploadRole=False , playlistRole=False , coverArtRole=False , 
+            commentRole=False , podcastRole=False , shareRole=False):
+        """
+        since 1.10.1
+
+        Modifies an existing Subsonic user.
+
+        username:str        The username of the user to update.
+        
+        All other args are the same as create user and you can update
+        whatever item you wish to update for the given username.
+
+        Returns a dict like the following:
+
+        {u'status': u'ok',
+         u'version': u'1.5.0',
+         u'xmlns': u'http://subsonic.org/restapi'}
+        """
+        methodName = 'updateUser'
+        viewName = '%s.view' % methodName
+        if password is not None:
+            password = 'enc:%s' % self._hexEnc(password)
+        q = self._getQueryDict({'username': username , 'password': password ,
+            'email': email , 'ldapAuthenticated': ldapAuthenticated , 
+            'adminRole': adminRole ,
+            'settingsRole': settingsRole , 'streamRole': streamRole ,
+            'jukeboxRole': jukeboxRole , 'downloadRole': downloadRole ,
+            'uploadRole': uploadRole , 'playlistRole': playlistRole ,
+            'coverArtRole': coverArtRole , 'commentRole': commentRole ,
+            'podcastRole': podcastRole , 'shareRole': shareRole})
+        req = self._getRequest(viewName , q)
+        res = self._doInfoReq(req)
+        self._checkStatus(res)
+        return res
+
     def deleteUser(self , username):
         """
         since: 1.3.0
@@ -962,6 +1053,8 @@ class Connection(object):
                         newest, highest, frequent, recent, 
                         (since 1.8.0 -> )starred, alphabeticalByName, 
                         alphabeticalByArtist
+                        Since 1.10.1 you can use byYear and byGenre to 
+                        list albums in a given year range or genre.
         size:int        The number of albums to return. Max 500
         offset:int      The list offset. Use for paging. Max 5000
 
@@ -1004,6 +1097,8 @@ class Connection(object):
                         newest, highest, frequent, recent, 
                         (since 1.8.0 -> )starred, alphabeticalByName, 
                         alphabeticalByArtist
+                        Since 1.10.1 you can use byYear and byGenre to 
+                        list albums in a given year range or genre.
         size:int        The number of albums to return. Max 500
         offset:int      The list offset. Use for paging. Max 5000
 
@@ -2049,11 +2144,56 @@ class Connection(object):
         self._checkStatus(res)
         return res
 
+    def scanMediaFolders(self):
+        """
+        This is not an officially supported method of the API
+
+        Same as selecting 'Settings' > 'Scan media folders now' with 
+        Subsonic web GUI
+
+        Returns True if refresh successful, False otherwise
+        """
+        methodName = 'scanNow'
+        return self._unsupportedAPIFunction(methodName)
+
+    def cleanupDatabase(self):
+        """
+        This is not an officially supported method of the API
+
+        Same as selecting 'Settings' > 'Clean-up Database' with Subsonic 
+        web GUI
+
+        Returns True if cleanup initiated successfully, False otherwise
+
+        Subsonic stores information about all media files ever encountered.
+        By cleaning up the database, information about files that are 
+        no longer in your media collection is permanently removed.
+        """
+        methodName = 'expunge'
+        return self._unsupportedAPIFunction(methodName)
+
+    def _unsupportedAPIFunction(self, methodName):
+        """
+        base function to call unsupported API methods
+
+        Returns True if refresh successful, False otherwise
+        :rtype : boolean
+        """
+        baseMethod = 'musicFolderSettings'
+        viewName = '%s.view' % baseMethod
+
+        url = '%s:%d/%s/%s?%s' % (self._baseUrl , self._port , 
+            self._separateServerPath() , viewName, methodName)
+        req = urllib2.Request(url)
+        res = self._opener.open(req)
+        res_msg = res.msg.lower()
+        return res_msg == 'ok'
+
     # Private internal methods
     def _getOpener(self , username , passwd):
         creds = b64encode('%s:%s' % (username , passwd))
         opener = urllib2.build_opener(PysHTTPRedirectHandler , 
-           HTTPSHandlerV3)
+            HTTPSHandlerChain)
         opener.addheaders = [('Authorization' , 'Basic %s' % creds)]
         return opener
 
@@ -2155,3 +2295,10 @@ class Connection(object):
         if ts is None:
             return None
         return int(ts * 1000)
+
+    def _separateServerPath(self):
+        """
+        separate REST portion of URL from base server path.
+        """
+        return urllib2.splithost(self._serverPath)[1].split('/')[0]
+
